@@ -1,6 +1,12 @@
+# -*- coding: utf-8 -*-
 import sys
 
 import six
+import typing
+
+
+if getattr(typing, 'TYPE_CHECKING', False):
+    from bravado_core._compat_typing import FuncType
 
 
 class SwaggerError(Exception):
@@ -23,7 +29,7 @@ class MatchingResponseNotFound(SwaggerMappingError):
 
 class SwaggerValidationError(SwaggerMappingError):
     """Raised when an error is encountered during validating user defined
-    format values in a request or a resposne.
+    format values in a request or a response.
     """
 
 
@@ -32,20 +38,31 @@ class SwaggerSchemaError(SwaggerError):
     """
 
 
+class SwaggerSecurityValidationError(SwaggerValidationError):
+    """Raised when an error is encountered during processing of
+    security related Swagger definitions
+    """
+
+
 def wrap_exception(exception_class):
+    # type: (typing.Type[BaseException]) -> typing.Callable[[FuncType], FuncType]
     """Helper decorator method to modify the raised exception class to
     `exception_class` but keeps the message and trace intact.
 
     :param exception_class: class to wrap raised exception with
     """
     def generic_exception(method):
+        # type: (FuncType) -> FuncType
         def wrapper(*args, **kwargs):
+            # type: (typing.Any, typing.Any) -> typing.Any
             try:
                 return method(*args, **kwargs)
             except Exception as e:
                 six.reraise(
                     exception_class,
                     exception_class(str(e)),
-                    sys.exc_info()[2])
-        return wrapper
+                    sys.exc_info()[2],
+                )
+        return wrapper  # type: ignore  # ignoring type to avoiding typing.cast call
+
     return generic_exception
