@@ -3,10 +3,13 @@ import abc
 import functools
 import logging
 import re
+from copy import deepcopy
 from warnings import warn
 
-import six
+import typing
+from six import add_metaclass
 from six import iteritems
+from six import string_types
 from swagger_spec_validator.ref_validators import attach_scope
 
 from bravado_core.schema import collapsed_properties
@@ -15,9 +18,15 @@ from bravado_core.schema import is_list_like
 from bravado_core.schema import is_ref
 from bravado_core.schema import SWAGGER_PRIMITIVES
 from bravado_core.util import determine_object_type
+from bravado_core.util import lazy_class_attribute
 from bravado_core.util import ObjectType
 from bravado_core.util import strip_xscope
 
+if getattr(typing, 'TYPE_CHECKING', False):
+    from bravado_core._compat_typing import JSONDict
+    from bravado_core.spec import Spec
+
+
 log = logging.getLogger(__name__)
 
 # Models in #/definitions are tagged with this key so that they can be
@@ -217,35 +226,52 @@ def _collect_models(container, json_reference, models, swagger_spec):
                     model_name=model_name,
                     model_type=model_type,
                     model_spec=model_spec,
-                    MODEL_MARKER=MODEL_MARKER
+                    MODEL_MARKER=MODEL_MARKER,
                 ),
             )
 
 
 class ModelMeta(abc.ABCMeta):
-    def __instancecheck__(self, instance):
-        """
-        An object is instance of a specific model class if:
-        - model class and instance share the same model Swagger Specs
-        - instance inherits from model class (in case of multiple inheritance, ie. allOff)
-        """
-        if not isinstance(type(instance), type(self)):
-            return False
+    def __instancecheck__(cls, instance):
+        return cls.__subclasscheck__(instance.__class__)
 
-        if not hasattr(self, '_model_spec'):
-            # This is the base Model class
-            return True
+    def __subclasscheck__(cls, subclass):
+        def _is_same_model(m1, m2):
+            """
+            Checks whether the two models are representing the same Model.
+            This method is needed to ensure that two different model type instances,
+            generated by the same swagger specs are considered equal.
+            """
+            try:
+                return m1._json_reference == m2._json_reference and m1._swagger_spec.origin_url == m2._swagger_spec.origin_url
+            except AttributeError:
+                return False
 
-        if not hasattr(instance, '_model_spec'):
-            return False
+        is_subclass = super(ModelMeta, cls).__subclasscheck__(subclass)
+        # NOTE: `and subclass` is added in order to short circuit issubclass check in case of
+        # `subclass is None`, this happens while the Python interpreter loads typing annotations
+        if not is_subclass and subclass:
+            try:
+                # Check whether cls is part any subclass._inherits_from
+                is_subclass = _is_same_model(cls, subclass) or any(
+                    _is_same_model(cls, value)
+                    for key, value in iteritems(subclass._swagger_spec.definitions)
+                    if key in subclass._inherits_from
+                )
+                if is_subclass:
+                    # The super method added subclass into _abc_negative_cache for faster lookup
+                    # As we determined that subclass is a valid subclass of cls we need to ensure
+                    # that the subclass is part of the already seen subclasses of cls
+                    cls._abc_negative_cache.remove(subclass)
+                    cls._abc_cache.add(subclass)
+            except AttributeError:
+                # Attribute error is possible if the passed subclass is not a `bravado_core.model.Model` class or subclass
+                pass
 
-        if self._model_spec == instance._model_spec:
-            return True
-
-        return self.__name__ in type(instance)._inherits_from
+        return is_subclass
 
 
-@six.add_metaclass(ModelMeta)
+@add_metaclass(ModelMeta)
 class Model(object):
     """Base class for Swagger models.
 
@@ -257,7 +283,7 @@ class Model(object):
     the names of attributes used in the Python implementation of the model
     (methods, etc.). The solution here is to have all non-property attributes
     making up the public API of this class prefixed by a single underscore
-    (this is done with the :func:`collections.namedtuple` type factory, which
+    (this is how :func:`collections.namedtuple` type factory works, which
     also uses property values with arbitrary names). There may still be name
     conflicts but only if the property name also begins with an underscore,
     which is uncommon. Truly private attributes are prefixed with double
@@ -279,14 +305,26 @@ class Model(object):
 
     .. attribute:: _model_spec
 
-        Class attribute that must be assigned on subclasses. JSON-like dict
-        that describes the model.
+        Class attribute that must be assigned on subclasses.
+        JSON-like dict that describes the model.
+
+    .. attribute:: _json_reference
+
+        Class attribute that must be assigned on subclasses.
+        JSON Uri where model spec could be found
 
     .. attribute:: _properties
 
-        Class attribute that must be assigned on subclasses. Dict mapping
-        property names to their specs. See
+        Class attribute that must be assigned on subclasses.
+        Dict mapping property names to their specs. See
         :func:`bravado_core.schema.collapsed_properties`.
+
+    .. attribute:: _inherits_from
+
+        Class attribute that must be assigned on subclasses.
+        List of the models from which the current model inherits from.
+        The list will be non-empty only for schemas with allOf
+
     """
 
     # Implementation details:
@@ -296,6 +334,15 @@ class Model(object):
     # __getattribute__ would have to have been overridden instead of
     # __getattr__.
 
+    # Use slots to reduce memory footprint of the Model instance
+    __slots__ = (
+        '_Model__dict',  # Note the name mangling!
+    )
+
+    _swagger_spec = None  # type: Spec
+    _model_spec = None  # type: JSONDict
+    _json_reference = None  # type: str
+
     def __init__(self, **kwargs):
         """Initialize from property values in keyword arguments.
 
@@ -303,12 +350,14 @@ class Model(object):
         """
         self.__init_from_dict(kwargs)
 
-    def __init_from_dict(self, dct, include_missing_properties=True):
+    def __init_from_dict(self, dct, include_missing_properties=None):
         """Initialize model from a dictionary of property values.
 
         :param dict dct: Dictionary of property values by name. They need not
             actually exist in :attr:`_properties`.
         """
+        if include_missing_properties is None:
+            include_missing_properties = self._swagger_spec.config['include_missing_properties']
 
         # Create the attribute value dictionary
         # We need bypass the overloaded __setattr__ method
@@ -318,10 +367,11 @@ class Model(object):
         # Additional property names in dct
         additional = set(dct).difference(self._properties)
 
-        if additional and not self._model_spec.get('additionalProperties', True):
+        if additional and self._model_spec.get('additionalProperties') is False:
             raise AttributeError(
-                "Model {0} does not have attributes for: {1}"
-                .format(type(self), list(additional))
+                "Model {0} does not have attributes for: {1}".format(
+                    type(self), list(additional),
+                ),
             )
 
         # Assign properties in model_spec, filling in None if missing from dct
@@ -333,6 +383,23 @@ class Model(object):
         for attr_name in additional:
             self.__dict[attr_name] = dct[attr_name]
 
+    @lazy_class_attribute
+    def _properties(self):
+        return collapsed_properties(self._model_spec, self._swagger_spec)
+
+    @lazy_class_attribute
+    def _inherits_from(self):
+        inherits_from_generator = (
+            _get_model_name(self._swagger_spec.deref(schema))
+            for schema in self._model_spec.get('allOf', [])
+        )
+
+        return [
+            inherits_from
+            for inherits_from in inherits_from_generator
+            if inherits_from is not None
+        ]
+
     def __contains__(self, obj):
         """Has a property set (including additional)."""
         return obj in self.__dict
@@ -350,8 +417,9 @@ class Model(object):
             return self[attr_name]
         except KeyError:
             raise AttributeError(
-                'type object {0!r} has no attribute {1!r}'
-                .format(type(self).__name__, attr_name)
+                'type object {0!r} has no attribute {1!r}'.format(
+                    type(self).__name__, attr_name,
+                ),
             )
 
     def __setattr__(self, attr_name, val):
@@ -404,8 +472,7 @@ class Model(object):
         Two model instances are equal if they have the same type and the same
         properties and values (including additional properties).
         """
-        # Check same type as self
-        if type(self) is not type(other):
+        if not isinstance(other, self.__class__):
             return False
 
         # Ignore any '_raw' keys
@@ -427,6 +494,12 @@ class Model(object):
         ]
         return "{0}({1})".format(self.__class__.__name__, ', '.join(s))
 
+    def __deepcopy__(self, memo=None):
+        """Deep copy all properties, but not metadata like the Swagger or Model spec attributes."""
+        if memo is None:  # pragma: no cover  # This should never happening, but better safe than sorry
+            memo = {}
+        return self.__class__(**deepcopy(self.__dict, memo=memo))
+
     @property
     def _additional_props(self):
         """Names of properties in instance which are not defined in spec."""
@@ -500,8 +573,8 @@ class Model(object):
 
         :rtype: dict
         """
-        from bravado_core.marshal import marshal_model
-        return marshal_model(self._swagger_spec, self._model_spec, self)
+        from bravado_core.marshal import marshal_schema_object
+        return marshal_schema_object(self._swagger_spec, self._model_spec, self)
 
     @classmethod
     def unmarshal(cls, val):
@@ -518,8 +591,8 @@ class Model(object):
         :type val: dict
         :rtype: .Model
         """
-        from bravado_core.unmarshal import unmarshal_model
-        return unmarshal_model(cls._swagger_spec, cls._model_spec, val)
+        from bravado_core.unmarshal import unmarshal_schema_object
+        return unmarshal_schema_object(cls._swagger_spec, cls._model_spec, val)
 
     @classmethod
     def isinstance(cls, obj):
@@ -556,8 +629,10 @@ class ModelDocstring(object):
 
     def __get__(self, obj, cls):
         if not hasattr(cls, '__docstring__'):
-            cls.__docstring__ = create_model_docstring(cls._swagger_spec,
-                                                       cls._model_spec)
+            cls.__docstring__ = create_model_docstring(
+                cls._swagger_spec,
+                cls._model_spec,
+            )
 
         return cls.__docstring__
 
@@ -581,21 +656,15 @@ def create_model_type(swagger_spec, model_name, model_spec, bases=(Model,), json
     :rtype: type
     """
 
-    inherits_from = []
-    if 'allOf' in model_spec:
-        for schema in model_spec['allOf']:
-            inherited_name = swagger_spec.deref(schema).get(MODEL_MARKER, None)
-            if inherited_name:
-                inherits_from.append(inherited_name)
-
-    return type(str(model_name), bases, dict(
-        __doc__=ModelDocstring(),
-        _swagger_spec=swagger_spec,
-        _model_spec=model_spec,
-        _properties=collapsed_properties(model_spec, swagger_spec),
-        _inherits_from=inherits_from,
-        _json_reference=json_reference,
-    ))
+    return type(
+        str(model_name), bases, dict(
+            __slots__=(),  # More memory-efficient
+            __doc__=ModelDocstring(),
+            _swagger_spec=swagger_spec,
+            _model_spec=model_spec,
+            _json_reference=json_reference,
+        ),
+    )
 
 
 def is_model(swagger_spec, schema_object_spec):
@@ -608,10 +677,15 @@ def is_model(swagger_spec, schema_object_spec):
     """
     deref = swagger_spec.deref
     schema_object_spec = deref(schema_object_spec)
-    return deref(schema_object_spec.get(MODEL_MARKER)) is not None
+    return isinstance(deref(schema_object_spec.get(MODEL_MARKER)), string_types)
 
 
-def is_object(swagger_spec, object_spec, no_default_type=False):
+def is_object(
+    swagger_spec,  # type: Spec
+    object_spec,  # type: JSONDict
+    no_default_type=False,  # type: bool
+):
+    # type: (...) -> bool
     """
     A schema definition is of type object if its type is object or if it uses
     model composition (i.e. it has an allOf property).
@@ -624,7 +698,8 @@ def is_object(swagger_spec, object_spec, no_default_type=False):
     """
     deref = swagger_spec.deref
     default_type = 'object' if not no_default_type and swagger_spec.config['default_type_to_object'] else None
-    return deref(object_spec.get('type', default_type)) == 'object' or 'allOf' in object_spec
+    object_type = deref(deref(object_spec).get('type', default_type))
+    return object_type == 'object' or (object_type is None and 'allOf' in object_spec)
 
 
 def create_model_docstring(swagger_spec, model_spec):
@@ -808,7 +883,7 @@ def _run_post_processing(spec):
         if uri == spec.origin_url or re.match(r'http://json-schema.org/draft-\d+/schema', uri)
     }
     additional_uri = _get_unprocessed_uri(spec, processed_uris)
-    while additional_uri:
+    while additional_uri is not None:
         # Post process each referenced specs to identify models in definitions of linked files
         with spec.resolver.in_scope(additional_uri):
             _call_post_process_spec(
@@ -848,3 +923,37 @@ def model_discovery(swagger_spec):
         # this ensures that the generated models have no references
         _run_post_processing(tmp_spec)
         swagger_spec.definitions = tmp_spec.definitions
+
+
+def _to_pickleable_representation(model_name, model_type):
+    # type: (typing.Text, typing.Type[Model]) -> typing.Dict[typing.Text, typing.Any]
+    """
+    Extract a pickleable representation of the input Model type.
+
+    Model types are runtime created types and so they are not pickleable.
+    In order to workaround this limitation we extract a representation,
+    which is pickleable such that we can re-create the input Model type
+    (via ``_from_pickleable_representation``).
+
+    NOTE:   This API should not be considered a public API and is meant
+            only to be used by bravado_core.spec.Spec.__getstate__ .
+    """
+    return {
+        'swagger_spec': model_type._swagger_spec,
+        'model_name': model_name,
+        'model_spec': model_type._model_spec,
+        'bases': model_type.__bases__,
+        'json_reference': model_type._json_reference,
+    }
+
+
+def _from_pickleable_representation(model_pickleable_representation):
+    # type: (typing.Dict[typing.Text, typing.Any]) -> typing.Type[Model]
+    """
+    Re-Create Model type form its pickleable representation
+    ``model_pickleable_representation`` is supposed to be the output of ``_to_pickleable_representation``.
+
+    NOTE:   This API should not be considered a public API and is meant
+            only to be used by bravado_core.spec.Spec.__getstate__ .
+    """
+    return create_model_type(**model_pickleable_representation)
