@@ -10,8 +10,11 @@ from six import iteritems
 from six import iterkeys
 from six import itervalues
 from six.moves.urllib.parse import ParseResult
+from six.moves.urllib.parse import urldefrag
 from six.moves.urllib.parse import urlparse
 from six.moves.urllib.parse import urlunparse
+from six.moves.urllib.request import pathname2url
+from six.moves.urllib.request import url2pathname
 from swagger_spec_validator.ref_validators import in_scope
 
 from bravado_core.model import model_discovery
@@ -83,7 +86,7 @@ def _marshal_uri(target_uri, origin_uri):
 
     if not marshalled_target or target_uri.scheme not in {'file', 'http', 'https'}:
         raise ValueError(
-            'Invalid target: \'{target_uri}\''.format(target_uri=urlunparse(target_uri))
+            'Invalid target: \'{target_uri}\''.format(target_uri=urlunparse(target_uri)),
         )
 
     if origin_uri and target_uri.scheme == 'file':
@@ -91,9 +94,9 @@ def _marshal_uri(target_uri, origin_uri):
 
         # Masquerade the absolute file path on the "local" server using
         # relative paths from the root swagger spec file
-        spec_dir = os.path.dirname(origin_uri.path)
+        spec_dir = os.path.dirname(url2pathname(origin_uri.path))
         scheme = 'lfile'
-        path = os.path.relpath(path, spec_dir)
+        path = pathname2url(os.path.relpath(url2pathname(path), spec_dir))
         marshalled_target = urlunparse((scheme, netloc, path, params, query, fragment))
 
     for src, dst in iteritems(MARSHAL_REPLACEMENT_PATTERNS):
@@ -218,7 +221,7 @@ class _SpecFlattener(object):
         def _rename_references_descend(value):
             if is_ref(value):
                 return {
-                    '$ref': reference_renaming_mapping.get(value['$ref'], value['$ref'])
+                    '$ref': reference_renaming_mapping.get(value['$ref'], value['$ref']),
                 }
             elif is_dict_like(value):
                 return {
@@ -280,7 +283,7 @@ class _SpecFlattener(object):
                     value == flattened_spec_dict.get('definitions', {}).get(value[MODEL_MARKER])
                 ):
                     return {
-                        '$ref': '#/definitions/{model_name}'.format(model_name=value[MODEL_MARKER])
+                        '$ref': '#/definitions/{model_name}'.format(model_name=value[MODEL_MARKER]),
                     }
 
                 else:
@@ -307,29 +310,71 @@ class _SpecFlattener(object):
         # Run model-discovery in order to tag the models available in known_mappings['definitions']
         # This is a required step that removes duplications of models due to the presence of models
         # in swagger.json#/definitions and the equivalent models generated by flattening
-        model_discovery(Spec(
-            spec_dict={
-                'definitions': {
-                    self.marshal_uri(uri): value
-                    for uri, value in iteritems(self.known_mappings['definitions'])
+        model_discovery(
+            Spec(
+                spec_dict={
+                    'definitions': {
+                        self.marshal_uri(uri): value
+                        for uri, value in iteritems(self.known_mappings['definitions'])
+                    },
                 },
-            },
-        ))
+            ),
+        )
 
-    def add_original_models_into_known_mappings(self):
-        flatten_models = {
-            # schema objects might not have a "type" set so they won't be tagged as models
-            definition.get(MODEL_MARKER)
-            for definition in itervalues(self.known_mappings['definitions'])
-        }
+    def include_discriminated_models(self):
+        """
+        This function ensures that discriminated models, present on the original Spec object
+        but not directly referenced by the flattened schema (because there is no direct $ref
+        attribute) are included in the final flattened specs
 
-        for model_name, model_type in iteritems(self.swagger_spec.definitions):
-            if model_name in flatten_models:
-                continue
-            model_url = urlparse(model_type._json_reference)
-            self.known_mappings['definitions'][model_url] = self.descend(
-                value=model_type._model_spec,
-            )
+        NOTE: The method re-run model_discovery in case additional models have been added to
+              the flattened models
+        """
+
+        def unflattened_models(flattened_models):
+            for m_name, m_type in iteritems(self.swagger_spec.definitions):
+                if m_name not in flattened_models:
+                    yield m_name, m_type
+
+        def register_unflattened_models():
+            """
+            :return: True if new models have been added
+            """
+            initial_number_of_models = len(self.known_mappings['definitions'])
+            modified = True
+
+            flattened_models = {
+                # schema objects might not have a "type" set so they won't be tagged as models
+                definition.get(MODEL_MARKER)
+                for definition in itervalues(self.known_mappings['definitions'])
+            }
+
+            while modified:
+                modified = False
+                for model_name, model_type in unflattened_models(flattened_models):
+                    if any(
+                        parent in flattened_models
+                        for parent in model_type._inherits_from
+                    ):
+                        model_url = urlparse(model_type._json_reference)
+                        flattened_models.add(model_name)
+                        self.known_mappings['definitions'][model_url] = self.descend(
+                            value=model_type._model_spec,
+                        )
+                        modified = True
+
+            return len(self.known_mappings['definitions']) != initial_number_of_models
+
+        while register_unflattened_models():
+            self.model_discovery()
+
+    def include_root_definition(self):
+        self.known_mappings['definitions'].update({
+            urlparse(v._json_reference): self.descend(value=v._model_spec)
+            for v in itervalues(self.swagger_spec.definitions)
+            # urldefrag(url)[0] returns the url without the fragment, it is guaranteed to be present
+            if urldefrag(v._json_reference)[0] == self.swagger_spec.origin_url
+        })
 
     @cached_property
     def resolved_specs(self):
@@ -339,10 +384,14 @@ class _SpecFlattener(object):
         # Perform model discovery of the newly identified definitions
         self.model_discovery()
 
+        # Ensure that all the root definitions, even if not referenced, are not lost due to flattening.
+        self.include_root_definition()
+
         # Add the identified models that are not available on the know_mappings definitions
-        # This could happen in case of models that have been discovered because on
-        # '#/definitions' of a referenced file, but not directly referenced by the specs
-        self.add_original_models_into_known_mappings()
+        # but that are related, via polymorphism (discriminator), to flattened models
+        # This could happen in case discriminated models are not directly referenced by the specs
+        # but is fair to assume that they should be on the final artifact due to polymorphism
+        self.include_discriminated_models()
 
         for mapping_key, mappings in iteritems(self.known_mappings):
             self.warn_if_uri_clash_on_same_marshaled_representation(mappings)
